from __future__ import annotations

import copy
import inspect
import json

from bidict import bidict
from contextlib import suppress
from typing import Any, Dict, Iterable, Iterator, List, Optional
from monday import MondayClient

from .helpers import as_type, as_obj, raise_monday_errors
from .fields import Field


class ItemMeta(type):
    __invalid_attribute_names__ = (
        # Attributes sets by the ItemMeta metaclass
        "_monday_client",
        "_field_names",
        "_board_id",
        "_monday_field_names",
        "_group_id",
        "_item_name",
        "_item_id",
        "_unsaved_item_name",
        # Properties from the Item class
        "fields",
        "board_id",
        "group_id",
        "item_name",
        "item_id",
    )

    def __new__(
        mcs,
        name,
        bases,
        attributes,
        monday_client: Optional[MondayClient] = None,
        board_id: Optional[int] = None,
        ignore_unused_fields: Optional[bool] = False,
    ):
        # Check if metaclass is running for class Item itself, in which case, it won't have any fields
        if not bases:
            return super().__new__(mcs, name, bases, attributes)

        # Validate that all of the attribute names are valid
        for field_name, field_obj in attributes.items():
            if field_name in ItemMeta.__invalid_attribute_names__:
                raise AttributeError(f"Attribute name {field_name} is invalid for an Item derive!")

        # If metaclass isn't running for the class `Item` itself
        # it must include `monday_client` and `board_id` as parameters
        if not monday_client:
            raise AttributeError(
                f"`monday_client` must be provided as parameter for Item derived class `{name}`"
            )

        if not board_id:
            raise AttributeError(f"`board_id` must be provided as parameter for Item derived class `{name}`")

        attributes["_board_id"] = board_id
        attributes["_monday_client"] = monday_client
        attributes["_frozen"] = False

        # Save all of the fields under the `_field_names` attribute
        # and create a backup of them under the _backup_fields
        attributes["_field_names"] = []
        attributes["_backup_fields"] = {}

        for field_name, field_obj in attributes.items():
            # Ignore attributes that arent fields
            if not issubclass(as_type(field_obj), Field):
                continue

            # Convert to object (this is for non-instansiated Fields)
            field_obj = as_obj(field_obj)
            attributes["_field_names"].append(field_name)
            attributes["_backup_fields"][field_name] = copy.deepcopy(field_obj)
            attributes[field_name] = field_obj

        # Iterate over all of the fields in the monday board and save the field names
        # and id's in a bidict, the reason I use a bidict is so it will be easy
        # to get the monday field id from the column name and vice versa.
        attributes["_monday_field_names"] = bidict()

        board_data = monday_client.boards.fetch_boards_by_id([board_id])

        # Check if the request succeed
        raise_monday_errors(board_data)

        for field_data in board_data["data"]["boards"][0]["columns"]:
            # Replace the field name from the monday board to be lowercase
            # and replace spaces into underscores
            # (so "My Example Column" will become "my_example_column")
            field_name = field_data["title"].lower().replace(" ", "_")

            # If we didn't found this field in our attributes just don't add it
            if field_name not in attributes:
                continue

            # Validate that the field type is the same as the expected field type
            # declared in the item class
            actual_field_type = field_data["type"].lower()
            expcted_field_type = attributes[field_name].__monday_field_type__
            if actual_field_type != expcted_field_type:
                raise AttributeError(
                    f"'{name}::{field_name}' should be of type '{expcted_field_type}' but got '{actual_field_type}'"
                )

            # Save the field name & id
            attributes["_monday_field_names"][field_name] = field_data["id"]

        # Iterate again over the item fields and check if there are
        # fields that aren't in use
        if not ignore_unused_fields:
            for field_name in attributes["_field_names"]:
                if field_name not in attributes["_monday_field_names"]:
                    raise AttributeError(f"'{name}::{field_name}' is declared but not used by the monday api")

        # Add some metadata for the monday fields to the attributes
        attributes["_item_name"] = None
        attributes["_item_id"] = None
        attributes["_group_id"] = None
        attributes["_unsaved_item_name"] = None
        attributes["_updated_fields"] = []

        return super().__new__(mcs, name, bases, attributes)


class Item(metaclass=ItemMeta):
    _field_names: List[str]
    _backup_fields: Dict[str, Field]
    _board_id: int
    _monday_client: MondayClient
    _monday_field_names: bidict
    _item_name: str
    _item_id: int
    _unsaved_item_name: str
    _group_id: str
    _frozen: bool

    @property
    def _changed_fields(self) -> Iterable[Field]:
        return (
            name for name in self._field_names if getattr(self, name).value != self._backup_fields[name].value
        )

    @property
    def _fields(self) -> Iterable[Field]:
        return (getattr(self, name) for name in self._field_names)

    @property
    def has_been_changed(self) -> bool:
        return any(self._changed_fields) or self._unsaved_item_name

    def __init__(self, *args, **kwargs):
        # Create a list of all the positional (required) arguments
        positional_args = []

        # if self is a subclass of Item (and not Item)
        if type(self) is not Item:
            for name, param in inspect.signature(self.__init__).parameters.items():
                # Check if the param is positional (required)
                if (
                    param.default == inspect.Parameter.empty
                    and param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                ):
                    positional_args.append(name)

        if len(args) != len(positional_args):
            raise ValueError(
                "Invalid arguments were passed to the superclass. "
                "Expected arguments: {} but {} given.".format(positional_args, len(args))
            )

        # Deepcopy the fields so different instances of Item have unique fields
        for name, field in self:
            setattr(self, name, copy.deepcopy(field))

        for k, v in kwargs.items():
            if k not in self._field_names:
                raise ValueError("Unexpected keyword argument given: {}={}".format(k, v))

            setattr(self, k, v)

        # Make sure no one can update this item attributes after the instanciation
        self._frozen = True

        super().__init__()

    def __str__(self):
        x = [
            f"\x1b[32m{self.__class__.__qualname__}\x1b[0m "
            f"(\x1b[32mId:\x1b[0m {self._item_id} | \x1b[32mName:\x1b[0m {self._item_name} | \x1b[32mGroup Id:\x1b[0m {self._group_id}):"
        ]

        for name, field in self:
            if isinstance(field, Item):
                x.append("\t{} ({}):".format(name, field.__class__.__qualname__))
                x.extend("\t{}".format(field_str) for field_str in str(field).splitlines()[1:])
            else:
                name_and_id = "\t{} <\x1b[33m{}\x1b[0m>: ".format(name, self._monday_field_names[name])

                value = "\x1b[33m{}\x1b[0m({})".format(field.__class__.__qualname__, str(field))

                x.append("{}{}".format(name_and_id.ljust(48), value))

        return "\n".join(x)

    def __eq__(self, other) -> bool:
        return all(a.value == b.value for a, b in zip(self._fields, other._fields))

    def __ne__(self, other) -> bool:
        return not self == other

    def __iter__(self):
        """
        :return: Iterator of (name, field) tuples
        """
        return zip(self._field_names, self._fields)

    def __setattr__(self, key, value):
        """
        Only allows setting the field's values.
        :param key:     The name of the attribute to set
        :param value:   The value to set
        :return:        None
        """

        if key in self._field_names and not isinstance(value, Field):
            # Update the value of the field and invoke the hooks
            field = getattr(self, key)
            field.value = value
            self.invoke_field_update_hooks(field)
        elif key in self._field_names:
            # Overriding the field so we must save the hooks attribute and set them
            # after we update the field
            field = getattr(self, key)
            hooks = getattr(field, "_value_update_hooks", [])
            super().__setattr__(key, value)
            setattr(field, "_value_update_hooks", hooks)

            # Invoke the field hooks because the value has been changed
            self.invoke_field_update_hooks(field)
        elif hasattr(self, key) or not self._frozen:
            super().__setattr__(key, value)
        else:
            raise AttributeError(
                f"Monday fields attributes can not be changed! ({self.__class__.__qualname__}:{key})"
            )

    @property
    def item_id(self):
        return self._item_id

    @property
    def item_name(self):
        return self._item_name

    @item_name.setter
    def item_name(self, value: str):
        self._unsaved_item_name = value

    @property
    def group_id(self):
        return self._group_id

    def duplicate_item(self) -> Item:
        if not self._item_id:
            raise AttributeError(f"Can not duplicate item that wasn't fetched from the server")

        item = copy.deepcopy(self)
        item._item_id = None
        item.create_item(self._group_id)
        return item

    def create_item(self, group_id: str):
        if self.item_id:
            raise AttributeError(f"The item already exists (id = {self._item_id})")

        # Get all the fields data
        column_values = {
            self._monday_field_names[field_name]: field.to_monday_dict()
            for field_name, field in self
            if field.to_monday_dict() is not None
        }

        # Create the item
        response = self._monday_client.items.create_item(
            self._board_id, group_id, self._unsaved_item_name, column_values, create_labels_if_missing=True
        )

        # Check if the request succeed
        raise_monday_errors(response)

        # Validate that the creation succeed
        self._item_id = response["data"]["create_item"]["id"]

        # Update the backup so it will hold those values now
        for field_name, field in self:
            self._backup_fields[field_name] = copy.deepcopy(field)

        # Set the needed attributes after the create request
        self._item_name = self._unsaved_item_name
        self._unsaved_item_name = None
        self._group_id = group_id

    def update_item(self):
        if not self.item_id:
            raise AttributeError(
                "Can not update an item that wasn't fetch from the server. Did you mean `create_item`?"
            )

        if not self.has_been_changed:
            # No need to update anything if there are no changes
            return

        # Get all the changed fields data (we wan't to update only what we change, not everything)
        column_values = {
            self._monday_field_names[field_name]: getattr(self, field_name).to_monday_dict()
            for field_name in self._changed_fields
            if getattr(self, field_name).to_monday_dict() is not None
        }

        if self._unsaved_item_name:
            column_values["name"] = self._unsaved_item_name

        # Update the item in the monday board
        data = self._monday_client.items.change_multiple_column_values(
            self._board_id, self.item_id, column_values, create_labels_if_missing=True
        )
        raise_monday_errors(data)

        # Update the name of the item
        if self._unsaved_item_name:
            self._item_name = self._unsaved_item_name
            self._unsaved_item_name = None

        # Update the backup items
        for field_name, field in self:
            self._backup_fields[field_name] = copy.deepcopy(field)

    def delete_item(self):
        if not self.item_id:
            raise AttributeError(
                "Can not update an item that wasn't fetch from the server. Did you mean `create_item`?"
            )

        data = self._monday_client.items.delete_item_by_id(self.item_id)

        # Check if the request succeed
        raise_monday_errors(data)

        self._unsaved_item_name = None
        self._item_id = None

    @classmethod
    def from_monday_dictionary(cls, data: Dict[str, Any]):
        obj = cls()
        obj._item_id = int(data["id"])
        obj._item_name = data["name"]
        obj._unsaved_item_name = None
        obj._group_id = data["group"]["id"]

        for column_data in data["column_values"]:
            monday_id = column_data["id"]

            # Check if the monday id is in our item dictionary
            if monday_id not in obj._monday_field_names.inverse:
                continue

            attribute_name = obj._monday_field_names.inverse[monday_id]

            column_value = None
            using_text = False

            with suppress(AttributeError):
                if getattr(obj, attribute_name).__use_text_instead_of_value__:
                    using_text = True
                    column_value = column_data["text"] if column_data["text"] else None

            # Load the data from the monday dictionary
            if not using_text:
                column_value = json.loads(column_data["value"]) if column_data["value"] else None

            getattr(obj, attribute_name).from_monday_dict(column_value)

            # Create a backup so we will know what have been changed
            obj._backup_fields[attribute_name] = copy.deepcopy(getattr(obj, attribute_name))

        return obj

    @classmethod
    def fetch_items_from_board(cls) -> Iterator[Item]:
        board_data = cls._monday_client.boards.fetch_items_by_board_id([cls._board_id])

        raise_monday_errors(board_data)

        for item in board_data["data"]["boards"][0]["items"]:
            yield cls.from_monday_dictionary(item)

    @classmethod
    def fetch_items_by_column_value(cls, **kwargs):
        field_name, field_value = next(iter(kwargs.items()))

        if field_name not in cls._monday_field_names:
            raise AttributeError("Invalid field to search by {}".format(field_name))

        monday_id = cls._monday_field_names[field_name]
        field_copy = copy.deepcopy(getattr(cls, field_name))
        field_copy.value = field_value
        data = field_copy.search_representation()

        items_data = cls._monday_client.items.fetch_items_by_column_value(cls._board_id, monday_id, data)

        # Check if the request succeed
        raise_monday_errors(items_data)

        for item in items_data["data"]["items_by_column_values"]:
            yield cls.from_monday_dictionary(item)

    @classmethod
    def fetch_group_ids(cls) -> Iterator[str]:
        groups_data = cls._monday_client.groups.get_groups_by_board([cls._board_id])

        # Check if the request succeed
        raise_monday_errors(groups_data)

        for group in groups_data["data"]["boards"][0]["groups"]:
            yield group["id"]

    def invoke_field_update_hooks(self, field: Field):
        for f in getattr(field, "_value_update_hooks", ()):
            f(self)

    @classmethod
    def field_updated_hook(cls, field):
        def register_field_hook(func: callable):
            if hasattr(field, "_value_update_hooks"):
                field._value_update_hooks.append(func)
            else:
                field._value_update_hooks = [func]

            return func

        return register_field_hook



from typing import Any, Dict, List

from .field import Field


class DropdownField(Field):
    __monday_field_type__ = "dropdown"

    def __init__(self, default: List[str] = []):
        self.value = default
        self._by = "labels"

    def to_monday_dict(self):
        if self._by == "ids":
            return {"ids": self._ids}

        return {"labels": self.value} if self.value else None

    def from_monday_dict(self, data: Dict[str, Any]):
        if data:
            self._by = "ids"
            self.value = [int(x) for x in data["ids"]]



from typing import Any, Dict, Union

from .field import Field


class StatusField(Field):
    __monday_field_type__ = "color"
    __type_to_field_name__ = {str: "label", int: "index"}

    def __init__(self, value: Union[str, int] = None):
        self.value = value

    def to_monday_dict(self):
        return (
            {StatusField.__type_to_field_name__[type(self.value)]: self.value}
            if self.value
            else None
        )

    def from_monday_dict(self, data: Dict[str, Any]):
        if not data:
            self.value = None
            return

        for key in StatusField.__type_to_field_name__.values():
            if key in data:
                self.value = data[key]
                break
        else:
            self.value = None

    def __str__(self):
        if self.value is None:
            return str(None)

        field_name = StatusField.__type_to_field_name__[type(self.value)]
        return f"{field_name}: {self.value}"

    def search_representation(self) -> str:
        if not self.value:
            raise ValueError("Can not search statuses by empty value")
        elif isinstance(self.value, int):
            raise TypeError("Can not search statuses by indexes, only by labels")

        return str(self.value)



from typing import List, Union, Dict

from .field import Field


class Person:
    def __init__(self, id: int):
        self.id = id

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    @property
    def kind(self):
        return "person"


class Team:
    def __init__(self, id: int):
        self.id = id

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    @property
    def kind(self):
        return "team"


class PeopleField(Field):
    __monday_field_type__ = "multiple-person"

    def __init__(self, default: List[Union[Person, Team]] = []):
        self.value = default

    def to_monday_dict(self):
        return (
            {"personsAndTeams": [{"id": x.id, "kind": x.kind} for x in self.value]}
            if self.value
            else {}
        )

    def from_monday_dict(self, data: Dict[str, str]):
        if not data:
            return

        value = []

        for people in data["personsAndTeams"]:
            if people["kind"] == "person":
                value.append(Person(people["id"]))
            elif people["kind"] == "team":
                value.append(Team(people["id"]))

        self.value = value

    def __str__(self):
        attrs = (
            ", ".join("{}: {}".format(x.kind, x.id) for x in self.value)
            if self.value
            else ""
        )

        return f"[{attrs}]"



from typing import Any, Dict

from .field import Field


class CheckboxField(Field):
    __monday_field_type__ = "boolean"

    def __init__(self, default: bool = False):
        self.value = default

    def to_monday_dict(self):
        return {"checked": "true"} if self.value else {}

    def from_monday_dict(self, data: Dict[str, Any]):
        self.value = data["checked"] if data else False



from collections import OrderedDict
from datetime import datetime
from typing import Dict

from .field import Field


class DateField(Field):
    __monday_field_type__ = "date"

    def __init__(self, default: datetime = None, include_time: bool = True):
        self.value = default
        self._include_time = include_time

    def to_monday_dict(self):
        data = (
            {
                "date": self.value.strftime("%Y-%m-%d"),
                "time": self.value.strftime("%H:%M:%S"),
            }
            if self.value
            else {}
        )

        if not self._include_time and "time" in data:
            del data["time"]

        return data

    def from_monday_dict(self, data: Dict[str, str]):
        if not data:
            self.value = None
            return

        variables = OrderedDict()
        time_format = OrderedDict()

        # Create the needed formatting only for the items that we have in the dictionary
        for key, time_fmt in {"date": "%Y-%m-%d", "time": "%H:%M:%S"}.items():
            if key not in data:
                continue

            variables[key] = data[key]
            time_format[key] = time_fmt

        # Parse the date & time by the formatting
        self.value = datetime.strptime(
            " ".join("{{{}}}".format(key) for key in variables).format(**variables),
            " ".join("{}".format(time_format[fmt]) for fmt in time_format),
        )

    def search_representation(self) -> str:
        return self.value.strftime("%Y-%m-%d")

    def __str__(self):
        format = "%Y-%m-%d"
        if self._include_time:
            format += " %H:%M:%S"

        return self.value.strftime(format) if self.value else str(None)



from .item import Item
from .exceptions import *
from .fields import __all__ as _fields_all
from .fields import *


field_updated_hook = Item.field_updated_hook

__all__ = ["Item", "MondayClientError", "field_updated_hook", *_fields_all]
__version__ = "0.1.0"



from .field import Field
from .checkbox import CheckboxField
from .country import CountryField
from .date import DateField
from .email import EmailField
from .link import LinkField, Link
from .number import NumberField
from .people import PeopleField, Person, Team
from .phone import PhoneField, Phone
from .status import StatusField
from .tags import TagsField
from .text import TextField, LongTextField
from .timeline import TimelineField, Timeline
from .mirror import MirrorField
from .dropdown import DropdownField
from .location import LocationField, Location, City, Country

__all__ = [
    "Field",
    "CheckboxField",
    "CountryField",
    "DateField",
    "EmailField",
    "LinkField", "Link",
    "NumberField",
    "PeopleField", "Person", "Team",
    "PhoneField", "Phone",
    "StatusField",
    "TagsField",
    "TextField",
    "LongTextField",
    "TimelineField", "Timeline",
    "MirrorField",
    "DropdownField",
    "LocationField", "Location", "City", "Country",
]



import abc

from typing import Any, Dict


class Field(abc.ABC):
    # This field is used to what is the matching type according
    # to Monday API when fetching data from monday boards
    __monday_field_type__ = None

    @abc.abstractmethod
    def to_monday_dict(self) -> Dict[str, Any]:
        """
        Represent the item as a monday dictionary.
        This dictionary is the way that Monday API knows how to create/update an item
        """

        raise NotImplementedError

    @abc.abstractmethod
    def from_monday_dict(self, data: Any):
        """
        Change the item value from the monday dictionary (fetch the data from the monday api)
        """

        raise NotImplementedError

    def search_representation(self) -> str:
        """
        Returns the representation of the value as it should be when we want to search it
        by the items by column values query
        (Read more at https://api.developer.monday.com/docs/items-by-column-values-queries)

        Only classes that will implement this function can be used by the `Item::fetch_items_by_column_value` function.
        """

        raise AttributeError(
            "items_by_column_value is not supported for {}".format(self.__class__.__qualname__)
        )

    @property
    def value(self) -> Any:
        """
        A property to get the current value of the field
        """

        return self._value

    @value.setter
    def value(self, value):
        """
        A property setter (set's the field value)
        """

        self._value = value

    def __str__(self) -> str:
        return str(self.value)



from typing import Dict


_COUNTRIES = {}


def get_available_countries() -> Dict[str, str]:
    """
    This function loads the list of countries from http://country.io to support
    monday's fields that need country codes (like Country and Phone fields).
    """

    global _COUNTRIES
    if not _COUNTRIES:
        import json
        import requests

        response = requests.get("http://country.io/names.json")
        if response.status_code != 200:
            raise Exception(
                "Failed to fetch countries, status_code={}, response={}".format(
                    response.status_code, response.text
                )
            )

        _COUNTRIES = json.loads(response.text)

    # Make sure we return a copy instead of the global variable
    # so it won't be changed
    return dict(_COUNTRIES)



import json

from typing import Union

from .field import Field


class NumberField(Field):
    __monday_field_type__ = "numeric"

    def __init__(self, default: Union[int, float] = None):
        self.value = default

    def to_monday_dict(self):
        return str(self.value) if self.value is not None else None

    def from_monday_dict(self, data: Union[int, float]):
        self.value = json.loads(data) if data else None

    def search_representation(self) -> str:
        return str(self.value)



from typing import Any, Dict

from .field import Field
from .helpers import get_available_countries


class CountryField(Field):
    __monday_field_type__ = "country"

    def __init__(self, default: str = None):
        self.value = default

    def to_monday_dict(self):
        return {
            "countryCode": self.value,
            "countryName": self.value,
        }

    def from_monday_dict(self, data: Dict[str, Any]):
        self.value = data["countryCode"] if data else None

    def search_representation(self) -> str:
        return str(self.value)

    @Field.value.setter
    def value(self, country_code: str):
        if not country_code:
            self._value = None
            return

        countries = get_available_countries()
        if country_code not in countries:
            raise ValueError(
                "Invalid country id, choose from ({})".format(", ".join(countries))
            )

        self._value = country_code



from typing import Any, Dict, List

from .field import Field


class TagsField(Field):
    __monday_field_type__ = "tag"

    def __init__(self, default: List[int] = []):
        self.value = default

    def to_monday_dict(self):
        return {"tag_ids": self.value}

    def from_monday_dict(self, data: Dict[str, Any]):
        self.value = data["tag_ids"] if data else None



from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase

from .field import Field
from ..helpers import remove_none_from_dict


@dataclass_json(letter_case=LetterCase.SNAKE)
@dataclass
class City:
    long_name: str | None = None
    short_name: str | None = None

    def from_monday_dict(self, data: dict[str, str]):
        self.long_name = data.get("long_name", None) if data else None
        self.short_name = data.get("short_name", None) if data else None


@dataclass_json(letter_case=LetterCase.SNAKE)
@dataclass
class Country:
    long_name: str | None = None
    short_name: str | None = None

    def from_monday_dict(self, data: dict[str, str]):
        self.long_name = data.get("long_name", None) if data else None
        self.short_name = data.get("short_name", None) if data else None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Location:
    address: str | None = None
    city: City | None = None
    country: Country | None = None
    lat: float | None = None
    lng: float | None = None
    place_id: str | None = None
    street: str | None = None
    street_short: str | None = None
    street_number: str | None = None
    street_number_short: str | None = None
    text: str | None = None

    def from_monday_dict(self, data: dict[str, str]):
        self.address = data.get("address", None) if data else None
        self.city = City().from_monday_dict(data.get("city", None)) if data else None
        self.country = Country().from_monday_dict(data.get("country", None)) if data else None
        self.lat = float(data.get("lat", None)) if data else None
        self.lng = float(data.get("lng", None)) if data else None
        self.place_id = data.get("placeId", None) if data else None
        self.street = data.get("street", None) if data else None
        self.street_short = data.get("streetShort", None) if data else None
        self.street_number = data.get("streetNumber", None) if data else None
        self.street_number_short = data.get("streetNumberShort", None) if data else None
        self.text = data.get("text", None) if data else None
        return self

    def to_monday_dict(self) -> dict[str, str]:
        return {
            "address": self.address,
            "lat": self.lat,
            "lng": self.lng,
        }
    
    def __str__(self):
        return self.text if self.text else "None"


class LocationField(Field):
    __monday_field_type__ = "location"

    def __init__(self, *args, **kwargs):
        self.value: Location = Location(*args, **kwargs)

    def to_monday_dict(self) -> dict[str, str] | None:
        clear_data = remove_none_from_dict(self.value.to_monday_dict()) if self.value else None
        print(clear_data)
        return clear_data or None

    def from_monday_dict(self, data: dict[str, str]):
        self.value = Location().from_monday_dict(data) 

    def __str__(self):
        return str(self.value.address) if self.value and self.value.address else "None"



from dataclasses import dataclass
from typing import Any, Dict

from .field import Field


@dataclass
class Link:
    url: str = None
    text: str = None

    def __str__(self):
        return "{} ({})".format(self.text, self.url)


class LinkField(Field):
    __monday_field_type__ = "link"

    def __init__(self, *args, **kwargs):
        self.value: Link = Link(*args, **kwargs)

    def to_monday_dict(self):
        return {"url": self.value.url, "text": self.value.text} if self.value else {}

    def from_monday_dict(self, data: Dict[str, Any]):
        self.value.url = data["url"] if data else None
        self.value.text = data["text"] if data else None

    def search_representation(self) -> str:
        return str(self.value.text)



from typing import Dict

from .field import Field


class TextField(Field):
    __monday_field_type__ = "text"

    def __init__(self, default: str = None):
        self.value = default

    def to_monday_dict(self):
        return self.value

    def from_monday_dict(self, data: str):
        self.value = data if data else None

    def search_representation(self) -> str:
        return str(self.value)


class LongTextField(Field):
    __monday_field_type__ = "long-text"

    def __init__(self, default: str = None):
        self.value = default

    def to_monday_dict(self):
        return {"text": self.value}

    def from_monday_dict(self, data: str):
        self.value = data if data else None

    def search_representation(self) -> str:
        return str(self.value)



import inspect

from .exceptions import MondayClientError


def as_type(obj):
    return obj if inspect.isclass(obj) else type(obj)


def as_obj(obj):
    return obj if not inspect.isclass(obj) else obj()


def remove_none_from_dict(data):
    return {k: v for k, v in data.items() if v is not None} if isinstance(data, dict) else data

def raise_monday_errors(response):
    if isinstance(response, dict) and "errors" in response:
        errors = [error["message"] if "message" in error else error for error in response["errors"]]

        raise MondayClientError("Got error from monday client", errors)



from dataclasses import dataclass
from datetime import datetime
from typing import Dict

from .field import Field


@dataclass
class Timeline:
    start: datetime = None
    end: datetime = None

    def __str__(self):
        return "{} -> {}".format(
            self.start.strftime("%Y-%m-%d"),
            self.end.strftime("%Y-%m-%d"),
        )


class TimelineField(Field):
    __monday_field_type__ = "timerange"

    def __init__(self, *args, **kwargs):
        self.value: Timeline = Timeline(*args, **kwargs)

    def to_monday_dict(self):
        return (
            {
                "from": self.value.start.strftime("%Y-%m-%d"),
                "to": self.value.end.strftime("%Y-%m-%d"),
            }
            if self.value and self.value.start and self.value.end
            else {}
        )

    def from_monday_dict(self, data: Dict[str, str]):
        if not data:
            self.value = None
            return

        self.value = Timeline(
            start=datetime.strptime(data["from"], "%Y-%m-%d"),
            end=datetime.strptime(data["to"], "%Y-%m-%d"),
        )

    def search_representation(self) -> str:
        return "{} - {}".format(
            self.value.start.strftime("%Y-%m-%d"),
            self.value.end.strftime("%Y-%m-%d"),
        )



from dataclasses import dataclass
from typing import Dict

from .field import Field
from .helpers import get_available_countries


@dataclass
class Phone:
    phone: str = None

    def __init__(self, **kwargs):
        self.phone = kwargs.pop("phone", None)
        self.country_code = kwargs.pop("country_code", None)

    @property
    def country_code(self) -> str:
        return self._country_code

    @country_code.setter
    def country_code(self, country_code: str):
        if not country_code:
            self._country_code = None
            return

        countries = get_available_countries()
        if country_code not in countries:
            raise ValueError(
                "Invalid country id, choose from ({})".format(", ".join(countries))
            )

        self._country_code = country_code

    def __str__(self) -> str:
        country_code = "({}) ".format(self.country_code) if self.country_code else ""
        return "{}{}".format(country_code, self.phone)


class PhoneField(Field):
    __monday_field_type__ = "phone"

    def __init__(self, *args, **kwargs):
        self.value: Phone = Phone(*args, **kwargs)

    def to_monday_dict(self):
        return (
            {
                "phone": self.value.phone,
                "countryShortName": self.value.country_code,
            }
            if self.value
            else {}
        )

    def from_monday_dict(self, data: Dict[str, str]):
        self.value.phone = data["phone"] if data else None
        self.value.country_code = data["countryShortName"] if data else None

    def search_representation(self) -> str:
        return str(self.value.phone)
